// Windows DIB/DirectDraw interface layer
// for the Build Engine
// by Jonathon Fowler (jonof@edgenetwk.com)
//
// Written for DirectX 6.0


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <ddraw.h>
#include <dinput.h>

#include "dxdidf.h"
#include <mmsystem.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#include "winlayer.h"
#include "pragmas.h"
#include "build.h"
#include "osd.h"


extern char moustat;

// undefine to restrict windowed resolutions to conventional sizes
#define ANY_WINDOWED_SIZE


// Windows crud
static HINSTANCE hInstance = 0;
static HWND hWindow = 0;
#define WINDOW_STYLE (WS_POPUP|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX)
extern long app_main(long argc, char *argv[]);
static BOOL window_class_registered = FALSE;

// DirectInput objects
static HMODULE              hDInputDLL    = NULL;
static LPDIRECTINPUTA       lpDI          = NULL;
static LPDIRECTINPUTDEVICEA lpDIDKeyboard = NULL;
static LPDIRECTINPUTDEVICEA lpDIDMouse    = NULL;
static BOOL                 bDInputInited = FALSE;
#define INPUT_BUFFER_SIZE	32
//#define USE_BUFFERED_KEYBOARD

// DirectDraw objects
static HMODULE              hDDrawDLL      = NULL;
static LPDIRECTDRAW         lpDD           = NULL;
static LPDIRECTDRAWSURFACE  lpDDSPrimary   = NULL;
static LPDIRECTDRAWSURFACE  lpDDSBack      = NULL;
static LPDIRECTDRAWSURFACE  lpDDSOffscreen = NULL;
static LPDIRECTDRAWPALETTE  lpDDPalette = NULL;
static BOOL                 bDDrawInited = FALSE;

// DIB stuff
static HDC      hDC         = NULL;
static HDC      hDCSection  = NULL;
static HBITMAP  hDIBSection = NULL;
static HPALETTE hPalette    = NULL;

#define NUM_SYS_COLOURS	25
static int syscolouridx[NUM_SYS_COLOURS] = {
	COLOR_SCROLLBAR,		// 1
	COLOR_BACKGROUND,
	COLOR_ACTIVECAPTION,
	COLOR_INACTIVECAPTION,
	COLOR_MENU,
	COLOR_WINDOW,
	COLOR_WINDOWFRAME,
	COLOR_MENUTEXT,
	COLOR_WINDOWTEXT,
	COLOR_CAPTIONTEXT,		// 10
	COLOR_ACTIVEBORDER,
	COLOR_INACTIVEBORDER,
	COLOR_APPWORKSPACE,
	COLOR_HIGHLIGHT,
	COLOR_HIGHLIGHTTEXT,
	COLOR_BTNFACE,
	COLOR_BTNSHADOW,
	COLOR_GRAYTEXT,
	COLOR_BTNTEXT,
	COLOR_INACTIVECAPTIONTEXT,	// 20
	COLOR_BTNHIGHLIGHT,
	COLOR_3DDKSHADOW,
	COLOR_3DLIGHT,
	COLOR_INFOTEXT,
	COLOR_INFOBK			// 25
};
static DWORD syscolours[NUM_SYS_COLOURS];
static char system_colours_saved = 0;

static const LPTSTR GetWindowsErrorMsg(DWORD code);
static const char * GetDDrawError(HRESULT code);
static const char * GetDInputError(HRESULT code);
static void ShowErrorBox(const char *m);
static void ShowDDrawErrorBox(const char *m, HRESULT r);
static void ShowDInputErrorBox(const char *m, HRESULT r);
static BOOL CheckWinVersion(void);
static LRESULT CALLBACK WndProcCallback(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static BOOL InitDirectDraw(void);
static void UninitDirectDraw(void);
static void ReleaseSurfaces(void);
static BOOL InitDirectInput(void);
static void UninitDirectInput(void);
static void AcquireInputDevices(char acquire);
static void ProcessInputDevices(void);
static void UninitDIB(void);
static BOOL RegisterWindowClass(void);
static BOOL CreateAppWindow(int width, int height, int fs);
static void DestroyAppWindow(void);
static void SaveSystemColours(void);
static void RestoreSystemColours(void);


// timer
static UINT_PTR timerid=0;
static void (*usertimercallback)(void) = NULL;
static int timerresolution=10;	//100 ticks/sec


// video
static long desktopxdim, desktopydim, desktoppal;
long xres=-1, yres=-1, fullscreen=0, bytesperline, imageSize;
long frameplace;
static PALETTEENTRY curPalette[256];
static int windowposx, windowposy;


// input and events
char quitevent=0, pauseevent=0, appactive=1, appminimized=0;
short mousex=0, mousey=0, mouseb=0, mousegrab=1;

char keystatus[256], keyfifo[KEYFIFOSIZ], keyfifoplc, keyfifoend;
static BYTE keys[2*256], keysofs=0;
static unsigned long lastKeyDown = 0;
static unsigned long lastKeyTime = 0;

#define SetKey(key,state) { \
	keystatus[key] = state; \
	keyfifo[keyfifoend] = key; \
	keyfifo[(keyfifoend+1)&(KEYFIFOSIZ-1)] = state; \
	keyfifoend = ((keyfifoend+2)&(KEYFIFOSIZ-1)); \
}





//-------------------------------------------------------------------------------------------------
//  MAIN CRAP
//=================================================================================================


//
// SignalHandler() -- called when we've sprung a leak
//
static void SignalHandler(int signum)
{
	switch (signum) {
		case SIGSEGV:
			printOSD("Fatal Signal caught: SIGSEGV. Bailing out.\n");
			uninitsystem();
			break;
		default:
			break;
	}
}


//
// WinMain() -- main Windows entry point
//
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	int r;
	char *argp;
	FILE *fp;
	int argc = 1;
	char *argv[30] = { "buildapp" };
	HDC hdc;

	hInstance = hInst;

	if (CheckWinVersion() || hPrevInst) {
		MessageBox(0, "This application must be run under Windows 95/98/Me or Windows 2000/XP or better.",
			apptitle, MB_OK|MB_ICONSTOP);
		return -1;
	}

	hdc = GetDC(NULL);
	r = GetDeviceCaps(hdc, BITSPIXEL);
	ReleaseDC(NULL, hdc);
	if (r < 8) {
		MessageBox(0, "This application requires a desktop colour depth of 256-colours or more.",
			apptitle, MB_OK|MB_ICONSTOP);
		return -1;
	}

	// carve up the commandline into more recognizable pieces
	argp = lpCmdLine;
	while (*argp && (argc < 30)) {
		while (((*argp <= 32) || (*argp > 126)) && (*argp != 0)) argp++;
		if (*argp == 0) break;

		argv[argc++] = argp;

		while ((*argp > 32) && (*argp <= 126)) argp++;
		if (*argp == 0) break;

		*(argp++) = 0;
	}

	// pipe standard outputs to files
	fp = freopen("stdout.txt", "w", stdout);
	if (!fp) {
		fp = fopen("stdout.txt", "w");
	}
	if (fp) setvbuf(fp, 0, _IONBF, 0);
	*stdout = *fp;

	// install signal handlers
	signal(SIGSEGV, SignalHandler);

	if (RegisterWindowClass()) return -1;

	MessageBox(0,
		"IMPORTANT:\n"
		"\tThis is a source port by Jonathon Fowler (jonof@edgenetwk.com) of the Build Engine, "
		"editor and test game by Ken Silverman to the Windows and Linux operating systems. It is "
		"distributed under the terms listed in BUILDLIC.TXT included with this package.",
		"Build Engine",
		MB_OK|MB_ICONINFORMATION);

	r = app_main(argc, argv);

	fclose(stdout);

	return r;
}


//
// initsystem() -- init systems
//
int initsystem(void)
{
	HDC hdc;
	int i;

	printOSD("Initializing Windows DirectX/GDI system interface\n");

	hdc = GetDC(NULL);

	// get the desktop dimensions before anything changes them
	desktopxdim = GetDeviceCaps(hdc, HORZRES);
	desktopydim = GetDeviceCaps(hdc, VERTRES);
	desktoppal  = ((GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE) == RC_PALETTE);

	ReleaseDC(NULL, hdc);

	if (desktoppal)
		// save the system colours
		SaveSystemColours();

	memset(curPalette, 0, sizeof(PALETTEENTRY) * 256);

	atexit(uninitsystem);

	// try and start DirectDraw
	if (InitDirectDraw())
		printOSD("DirectDraw initialization failed. Fullscreen modes will be unavailable.\n");

	return 0;
}


//
// uninitsystem() -- uninit systems
//
void uninitsystem(void)
{
	DestroyAppWindow();

	uninitinput();
	uninittimer();
}


//
// handleevents() -- process the Windows message queue
//   returns !0 if there was an important event worth checking (like quitting)
//
int handleevents(void)
{
	int rv=0;
	MSG msg;

	//if (appminimized) WaitMessage();

	while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
		if (msg.message == WM_QUIT)
			quitevent = 1;
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	if (!appactive || quitevent) rv = -1;

	return rv;
}







//-------------------------------------------------------------------------------------------------
//  INPUT (MOUSE/KEYBOARD/JOYSTICK?)
//=================================================================================================

//
// initinput() -- init input system
//
int initinput(void)
{
	moustat=0;
	memset(keystatus, 0, sizeof(keystatus));
	memset(keys, 0, sizeof(keys));
	keysofs = 0;

	if (InitDirectInput())
		return -1;

	AcquireInputDevices(1);

	return 0;
}


//
// uninitinput() -- uninit input system
//
void uninitinput(void)
{
	uninitmouse();
	UninitDirectInput();
}


//
// initmouse() -- init mouse input
//
int initmouse(void)
{
	if (moustat) return 0;

	printOSD("Initializing mouse\n");

	// grab input
	moustat=1;
	grabmouse(1);

	return 0;
}


//
// uninitmouse() -- uninit mouse input
//
void uninitmouse(void)
{
	if (!moustat) return;

	grabmouse(0);
	moustat=0;
}


//
// grabmouse() -- show/hide mouse cursor
//
void grabmouse(char a)
{
	if (!moustat) return;

	mousegrab = a;

	AcquireInputDevices(a);

	mousex = 0;
	mousey = 0;
	mouseb = 0;
}


//
// readmousexy() -- return mouse motion information
//
void readmousexy(short *x, short *y)
{
	*x = mousex;
	*y = mousey;
	mousex = 0;
	mousey = 0;
}


//
// readmousebstatus() -- return mouse button information
//
void readmousebstatus(short *b)
{
	*b = mouseb;
}







//-------------------------------------------------------------------------------------------------
//  TIMER
//=================================================================================================

//
// timercallback() -- multimedia timer callback
//
static void CALLBACK timercallback(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	totalclock++;
	ProcessInputDevices();
	if (usertimercallback) usertimercallback();
	//handleevents();
}


//
// inittimer() -- install timer callback
//
int inittimer(int tickspersecond)
{
	TIMECAPS tc;
	
	if (timerid) return 0;	// already installed

	printOSD("Initializing timer\n");

	if (timeGetDevCaps(&tc, sizeof(tc)) != TIMERR_NOERROR) {
		ShowErrorBox("Failed calling timeGetDevCaps()");
		return -1;
	}

	timerresolution = 1000 / tickspersecond;
	if (timerresolution < tc.wPeriodMin) {
		timerresolution = tc.wPeriodMin;
		printOSD("Warning: %d tickspersecond is too fast for this system. Falling back to minimum of %d\n", tickspersecond, 1000/timerresolution);
	}

	if (timeBeginPeriod(timerresolution) != TIMERR_NOERROR) {
		ShowErrorBox("Failed calling timeBeginPeriod()");
		return -1;
	}

	usertimercallback = NULL;

	halttimer(0);

	return 0;
}

//
// uninittimer() -- remove timer callback
//
void uninittimer(void)
{
	if (!timerid) return;

	halttimer(1);

	timeEndPeriod(timerresolution);

	timerid=0;
}

//
// halttimer() -- freezes and thaws time
//
int halttimer(char halt)
{
	if (halt && !timerid) return 0;		// already stopped
	else if (!halt && timerid) return 0;	// already running

	if (halt) {
		timeKillEvent(timerid);
		timerid = 0;
	} else {
		timerid = timeSetEvent(timerresolution, timerresolution/2, timercallback, 0, TIME_PERIODIC);
		if (!timerid) {
			printOSD("Timer initialization failed: %s\n", GetWindowsErrorMsg(GetLastError()));
			return -1;	// error
		}
	}

	return 0;
}

//
// getticks() -- returns the windows ticks count
//
long getticks(void)
{
	return (long)timeGetTime();
}


//
// installusertimercallback() -- set up a callback function to be called when the timer is fired
//
void (*installusertimercallback(void (*callback)(void)))(void)
{
	void (*oldtimercallback)(void);

	oldtimercallback = usertimercallback;
	usertimercallback = callback;

	return oldtimercallback;
}







//-------------------------------------------------------------------------------------------------
//  VIDEO
//=================================================================================================

//
// checkvideomode() -- makes sure the video mode passed is legal
//
int checkvideomode(long *x, long *y, int fs)
{
	int i, nearest=-1, dx, dy, odx=9999, ody=9999;

	getvalidmodes();

	// fix up the passed resolution values to be multiples of 8
	// and at least 320x200 or at most MAXXDIMxMAXYDIM
	if (*x < 320) *x = 320;
	if (*y < 200) *y = 200;
	if (*x > MAXXDIM) *x = MAXXDIM;
	if (*y > MAXYDIM) *y = MAXYDIM;
	*x &= 0xfffffff8l;

#ifdef ANY_WINDOWED_SIZE
	if (fs) {
#endif
		for (i=0; i<validmodecnt; i++) {
			if (validmodefs[i] != fs) continue;
			dx = klabs(validmodexdim[i] - *x);
			dy = klabs(validmodeydim[i] - *y);
			if (!(dx | dy)) { 	// perfect match
				nearest = i;
				break;
			}
			if ((dx <= odx) && (dy <= ody)) {
				nearest = i;
				odx = dx; ody = dy;
			}
		}

		if (nearest < 0) {
			// no mode that will match (eg. if no fullscreen modes)
			return -1;
		}
		*x = validmodexdim[nearest];
		*y = validmodeydim[nearest];
#ifdef ANY_WINDOWED_SIZE
	}
#endif

	return 0;
}


//
// setvideomode() -- set the video mode
//
int setvideomode(int x, int y, int fs)
{
	char t[384];
	HRESULT result;
	DDSURFACEDESC ddsd;

	if ((fs == fullscreen) && (x == xres) && (x == yres)) return 0;

	halttimer(1);
	AcquireInputDevices(0);

	printOSD("Setting video mode %dx%d (%s)\n", x,y, ((fs) ? "fullscreen" : "windowed"));
	if (CreateAppWindow(x, y, fs)) return -1;

	sprintf(t, "%s (%dx%d %s)", apptitle, x, y, ((fs) ? "fullscreen" : "windowed"));
	SetWindowText(hWindow, t);

	AcquireInputDevices(1);
	halttimer(0);

	return 0;
}


//
// getvalidmodes() -- figure out what video modes are available
//
#define ADDMODE(x,y,f) { \
	validmodexdim[validmodecnt]=x; \
	validmodeydim[validmodecnt]=y; \
	validmodefs[validmodecnt]=f; \
	validmodecnt++; \
	printOSD("Adding mode %dx%d (%s)\n", x, y, (f)?"fullscreen":"windowed"); }

#define CHECK(w,h) if ((w <= maxx) && (h <= maxy))

// mode enumerator
static HRESULT WINAPI getvalidmodes_enum(DDSURFACEDESC *ddsd, VOID *udata)
{
	unsigned maxx = MAXXDIM, maxy = MAXYDIM;

	if (ddsd->ddpfPixelFormat.dwRGBBitCount == 8) {
		CHECK(ddsd->dwWidth, ddsd->dwHeight) ADDMODE(ddsd->dwWidth, ddsd->dwHeight, 1);
	}

	return(DDENUMRET_OK);
}

void getvalidmodes(void)
{
	static int modeschecked=0;
	int i, maxx=0, maxy=0;
	HRESULT result;

	if (modeschecked) return;

	validmodecnt=0;

	if (bDDrawInited) {
		// if DirectDraw initialization didn't fail enumerate fullscreen modes

		result = IDirectDraw_EnumDisplayModes(lpDD, 0, NULL, 0, getvalidmodes_enum);
		if (result != DD_OK) {
			printOSD("Unable to enumerate fullscreen modes. Using default list.\n");
			ADDMODE(1024,768,1)
			ADDMODE(800,600,1)
			ADDMODE(640,480,1)
			ADDMODE(640,400,1)
			ADDMODE(512,384,1)
			ADDMODE(480,360,1)
			ADDMODE(400,300,1)
			ADDMODE(320,240,1)
			ADDMODE(320,200,1)
		}
	}

	// windowed modes cant be bigger than the current desktop resolution
	maxx = desktopxdim-1;
	maxy = desktopydim-1;

	// add windowed modes next
	CHECK(320,200) ADDMODE(320,200,0)
	CHECK(320,240) ADDMODE(320,240,0)
	CHECK(400,300) ADDMODE(400,300,0)
	CHECK(480,360) ADDMODE(480,360,0)
	CHECK(512,384) ADDMODE(512,384,0)
	CHECK(640,400) ADDMODE(640,400,0)
	CHECK(640,480) ADDMODE(640,480,0)
	CHECK(800,600) ADDMODE(800,600,0)
	CHECK(1024,768) ADDMODE(1024,768,0)
	CHECK(1152,864) ADDMODE(1152,864,0)
	CHECK(1280,960) ADDMODE(1280,960,0)
	CHECK(1280,1024) ADDMODE(1280,1024,0)

	modeschecked=1;
}

#undef CHECK
#undef ADDMODE


//
// showframe() -- update the display
//
void showframe(void)
{
	HRESULT result;
	DDSURFACEDESC ddsd;
	RECT rect;

	if (!fullscreen) {
		BitBlt(hDC, 0, 0, xres, yres, hDCSection, 0, 0, SRCCOPY);
	} else {
		result = IDirectDrawSurface_Unlock(lpDDSOffscreen, NULL);
		if (result != DD_OK) {
			printOSD("Failed unlocking offscreen surface: %s\n", GetDDrawError(result));
			return;
		}

		rect.left = 0;
		rect.top = 0;
		rect.right = xres - 1;
		rect.bottom = yres - 1;

		result = IDirectDrawSurface_BltFast(lpDDSBack, 0, 0, lpDDSOffscreen, &rect, /*DDBLTFAST_WAIT*/0);
		if (result == DDERR_SURFACELOST) {
			IDirectDrawSurface_Restore(lpDDSBack);
			result = IDirectDrawSurface_BltFast(lpDDSBack, 0, 0, lpDDSOffscreen, &rect, /*DDBLTFAST_WAIT*/0);
		}

		if (result == DD_OK) {
			result = IDirectDrawSurface_Flip(lpDDSPrimary, NULL, 0);
			if (result == DDERR_SURFACELOST) {
				IDirectDrawSurface_Restore(lpDDSPrimary);
				result = IDirectDrawSurface_Flip(lpDDSPrimary, NULL, 0);
			}
		}

		memset(&ddsd, 0, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);

		result = IDirectDrawSurface_Lock(lpDDSOffscreen, NULL, &ddsd, DDLOCK_NOSYSLOCK | DDLOCK_WAIT, NULL);
		if (result != DD_OK) {
			printOSD("Failed locking offscreen surface: %s\n", GetDDrawError(result));
			return;
		}

		frameplace = (long)ddsd.lpSurface;
	}
}


//
// setpalette() -- set palette values
//
int setpalette(int start, int num, char *dapal)
{
	int i, n;
	HRESULT result;
	RGBQUAD *rgb;
	HPALETTE hPalPrev;

	struct logpal {
		WORD palVersion;
		WORD palNumEntries;
		PALETTEENTRY palPalEntry[256];
	} *lpal;

	for (i=start, n=num; n>0; i++, n--) {
		curPalette[i].peBlue = dapal[0] << 2;
		curPalette[i].peGreen = dapal[1] << 2;
		curPalette[i].peRed = dapal[2] << 2;
		curPalette[i].peFlags = PC_RESERVED | PC_NOCOLLAPSE;
		dapal += 4;
	}

	if (!fullscreen) {
		if (num > 0) {
			rgb = (RGBQUAD *)alloca(sizeof(RGBQUAD)*num);
			for (i=start, n=0; n<num; i++, n++) {
				rgb[n].rgbBlue = curPalette[i].peBlue;
				rgb[n].rgbGreen = curPalette[i].peGreen;
				rgb[n].rgbRed = curPalette[i].peRed;
				rgb[n].rgbReserved = 0;
			}

			SetDIBColorTable(hDCSection, start, num, rgb);
		}

		if (!desktoppal) return 0;	// only if an 8bit desktop do we do what follows

		lpal = (struct logpal *)alloca(sizeof(struct logpal));

		memcpy(lpal->palPalEntry, curPalette, sizeof(PALETTEENTRY)*256);

		// set 0 and 255 to black and white
		lpal->palVersion = 0x300;
		lpal->palNumEntries = 256;
		lpal->palPalEntry[0].peBlue = 0;
		lpal->palPalEntry[0].peGreen = 0;
		lpal->palPalEntry[0].peRed = 0;
		lpal->palPalEntry[0].peFlags = 0;
		lpal->palPalEntry[255].peBlue = 255;
		lpal->palPalEntry[255].peGreen = 255;
		lpal->palPalEntry[255].peRed = 255;
		lpal->palPalEntry[255].peFlags = 0;

		if (SetSystemPaletteUse(hDC, SYSPAL_NOSTATIC) == SYSPAL_ERROR) {
			//printOSD("Problem setting system palette use: %s\n", GetWindowsErrorMsg(GetLastError()));
			return -1;
		}

		if (hPalette) {
			if (num == 0) { start = 0; num = 256; }		// refreshing the palette only
			SetPaletteEntries(hPalette, start, num, lpal->palPalEntry+start);
		} else {
			hPalette = CreatePalette((LOGPALETTE *)lpal);
			if (!hPalette) {
				printOSD("Problem creating palette: %s\n", GetWindowsErrorMsg(GetLastError()));
				return -1;
			}
		}

		if (SelectPalette(hDC, hPalette, FALSE) == NULL) {
			printOSD("Problem selecting palette: %s\n", GetWindowsErrorMsg(GetLastError()));
			return -1;
		}

		if (RealizePalette(hDC) == GDI_ERROR) {
			printOSD("Failure realizing palette: %s\n", GetWindowsErrorMsg(GetLastError()));
			return -1;
		}
	} else {
		result = IDirectDrawPalette_SetEntries(lpDDPalette, 0, start, num, &curPalette[start]);
		if (result != DD_OK) {
			printOSD("Palette set failed: %s\n", GetDDrawError(result));
			return -1;
		}
	}

	return 0;
}

//
// getpalette() -- get palette values
//
int getpalette(int start, int num, char *dapal)
{
	int i;

	for (i=num; i>0; i--, start++) {
		dapal[0] = curPalette[start].peBlue >> 2;
		dapal[1] = curPalette[start].peGreen >> 2;
		dapal[2] = curPalette[start].peRed >> 2;
		dapal += 4;
	}

	return 0;
}






//-------------------------------------------------------------------------------------------------
//  MOSTLY STATIC INTERNAL WINDOWS THINGS
//=================================================================================================

//
// ShowErrorBox() -- shows an error message box
//
static void ShowErrorBox(const char *m)
{
	TCHAR msg[1024];

	wsprintf(msg, "%s: %s", m, GetWindowsErrorMsg(GetLastError()));
	MessageBox(0, msg, apptitle, MB_OK|MB_ICONSTOP);
}


//
// ShowDDrawErrorBox() -- shows an error message box for a DirectDraw error
//
static void ShowDDrawErrorBox(const char *m, HRESULT r)
{
	TCHAR msg[1024];

	wsprintf(msg, "%s: %s", m, GetDDrawError(r));
	MessageBox(0, msg, apptitle, MB_OK|MB_ICONSTOP);
}


//
// ShowDInputErrorBox() -- shows an error message box for a DirectInput error
//
static void ShowDInputErrorBox(const char *m, HRESULT r)
{
	TCHAR msg[1024];

	wsprintf(msg, "%s: %s", m, GetDInputError(r));
	MessageBox(0, msg, apptitle, MB_OK|MB_ICONSTOP);
}


//
// CheckWinVersion() -- check to see what version of Windows we happen to be running under
//
static BOOL CheckWinVersion(void)
{
	OSVERSIONINFO osv;

	osv.dwOSVersionInfoSize = sizeof(osv);
	if (!GetVersionEx(&osv)) return TRUE;

	// haha, yeah, like it will work on Win32s
	if (osv.dwPlatformId == VER_PLATFORM_WIN32s) return TRUE;

	// we don't like NT 3.51
	if (osv.dwMajorVersion < 4) return TRUE;

	// nor do we like NT 4
	if (osv.dwPlatformId == VER_PLATFORM_WIN32_NT &&
	    osv.dwMajorVersion == 4) return TRUE;

	return FALSE;
}


//
// WndProcCallback() -- the Windows window callback
//
static LRESULT CALLBACK WndProcCallback(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	RECT rect;
	POINT pt;
	HRESULT result;

	switch (uMsg) {
		case WM_ACTIVATE:
			appminimized = HIWORD(wParam);
			appactive = (LOWORD(wParam) != WA_INACTIVE);

			if (desktoppal) {
				if (appactive) {
					setpalette(0,0,0);
				} else {
					RestoreSystemColours();
				}
			}

			AcquireInputDevices(appactive && !appminimized);
			break;

		case WM_ENTERMENULOOP:
		case WM_ENTERSIZEMOVE:
			// un-acquire device when entering menu or re-sizing
			// this will show the mouse cursor again
			appactive = 0;
			AcquireInputDevices(0);
			//halttimer(1);
			break;

		case WM_EXITMENULOOP:
		case WM_EXITSIZEMOVE:
			// re-acquire device when leaving menu or re-sizing
			// this will show the mouse cursor again

			// even though the menu is going away, the app
			// might have lost focus or be an icon
			if (GetActiveWindow() == hWnd && !IsIconic(hWnd))
				appactive = 1;
			else
				appactive = 0;

			//halttimer(0);
			AcquireInputDevices(appactive);
			break;

		case WM_PAINT:
			// blit the back-buffer to the screen
			break;

		case WM_MOVE:
			windowposx = LOWORD(lParam);
			windowposy = HIWORD(lParam);
			return 0;

		case WM_CLOSE:
			quitevent = 1;
			return 0;

		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
			return 0;

		case WM_SYSCOMMAND:
			if (wParam == SC_SCREENSAVE) return 0;
			break;

		case WM_DESTROY:
			hWindow = 0;
			PostQuitMessage(0);
			return 0;

		default:
			break;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


//
// RegisterWindowClass() -- register the window class
//
static BOOL RegisterWindowClass(void)
{
	WNDCLASSEX wcx;

	if (window_class_registered) return FALSE;

	printOSD("Registering window class\n");

	wcx.cbSize	= sizeof(wcx);
	wcx.style	= 0;
	wcx.lpfnWndProc	= WndProcCallback;
	wcx.cbClsExtra	= 0;
	wcx.cbWndExtra	= 0;
	wcx.hInstance	= hInstance;
	wcx.hIcon	= LoadImage(hInstance, MAKEINTRESOURCE(100), IMAGE_ICON,
				GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), LR_DEFAULTCOLOR);
	wcx.hCursor	= LoadCursor(hInstance, IDC_ARROW);
	wcx.hbrBackground = (HBRUSH)COLOR_GRAYTEXT;
	wcx.lpszMenuName = NULL;
	wcx.lpszClassName = "buildapp";
	wcx.hIconSm	= LoadImage(hInstance, MAKEINTRESOURCE(100), IMAGE_ICON,
				GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), LR_DEFAULTCOLOR);
	if (!RegisterClassEx(&wcx)) {
		ShowErrorBox("Failed to register window class");
		return TRUE;
	}

	window_class_registered = TRUE;

	return FALSE;
}


//
// InitDirectDraw() -- get DirectDraw started
//
static BOOL InitDirectDraw(void)
{
	HRESULT result;
	HRESULT (WINAPI *aDirectDrawCreate)(GUID *, LPDIRECTDRAW *, IUnknown *);

	if (bDDrawInited) return FALSE;

	printOSD("Initializing DirectDraw...\n");

	// load up the DirectDraw DLL
	if (!hDDrawDLL) {
		printOSD("  - Loading DDRAW.DLL\n");
		hDDrawDLL = LoadLibrary("DDRAW.DLL");
		if (!hDDrawDLL) {
			ShowErrorBox("Error loading DDRAW.DLL");
			return TRUE;
		}
	}

	// get the pointer to DirectDrawCreate
	aDirectDrawCreate = (void *)GetProcAddress(hDDrawDLL, "DirectDrawCreate");
	if (!aDirectDrawCreate) {
		ShowErrorBox("Error fetching DirectDrawCreate()");
		UninitDirectDraw();
		return TRUE;
	}

	// create a new DirectDraw object
	printOSD("  - Creating DirectDraw object\n");
	result = aDirectDrawCreate(NULL, &lpDD, NULL);
	if (result != DD_OK) {
		ShowDDrawErrorBox("DirectDrawCreate() failed", result);
		UninitDirectDraw();
		return TRUE;
	}

	bDDrawInited = TRUE;

	return FALSE;
}


//
// UninitDirectDraw() -- clean up DirectDraw
//
static void UninitDirectDraw(void)
{
	if (bDDrawInited) printOSD("Uninitializing DirectDraw...\n");

	AcquireInputDevices(0);
	ReleaseSurfaces();

	if (lpDD) {
		printOSD("  - Releasing DirectDraw object\n");
		IDirectDraw_Release(lpDD);
		lpDD = NULL;
	}

	if (hDDrawDLL) {
		printOSD("  - Unloading DDRAW.DLL\n");
		FreeLibrary(hDDrawDLL);
		hDDrawDLL = NULL;
	}

	bDDrawInited = FALSE;
}


//
// ReleaseSurfaces() -- release the front and back buffers
//
static void ReleaseSurfaces(void)
{
	if (lpDDPalette) {
		printOSD("  - Releasing palette\n");
		IDirectDrawPalette_Release(lpDDPalette);
		lpDDPalette = NULL;
	}

	if (lpDDSPrimary) {
		printOSD("  - Releasing primary surface\n");
		IDirectDrawSurface_Release(lpDDSPrimary);
		lpDDSPrimary = NULL;
	}

	if (lpDDSOffscreen) {
		printOSD("  - Releasing offscreen surface\n");
		IDirectDrawSurface_Release(lpDDSOffscreen);
		lpDDSOffscreen = NULL;
	}
}


//
// InitDirectInput() -- get DirectInput started
//
static BOOL InitDirectInput(void)
{
	HRESULT result;
	HRESULT (WINAPI *aDirectInputCreateA)(HINSTANCE, DWORD, LPDIRECTINPUTA *, LPUNKNOWN);
	DIPROPDWORD dipdw;

	if (bDInputInited) return FALSE;

	printOSD("Initializing DirectInput...\n");

	// load up the DirectInput DLL
	if (!hDInputDLL) {
		printOSD("  - Loading DINPUT.DLL\n");
		hDInputDLL = LoadLibrary("DINPUT.DLL");
		if (!hDInputDLL) {
			ShowErrorBox("Error loading DINPUT.DLL");
			return TRUE;
		}
	}

	// get the pointer to DirectInputCreate
	aDirectInputCreateA = (void *)GetProcAddress(hDInputDLL, "DirectInputCreateA");
	if (!aDirectInputCreateA) {
		ShowErrorBox("Error fetching DirectInputCreateA()");
		UninitDirectInput();
		return TRUE;
	}

	// create a new DirectInput object
	printOSD("  - Creating DirectDraw object\n");
	result = aDirectInputCreateA(hInstance, DIRECTINPUT_VERSION, &lpDI, NULL);
	if (result != DI_OK) {
		ShowDInputErrorBox("DirectInputCreateA() failed", result);
		UninitDirectInput();
		return TRUE;
	}

	// create the keyboard device
	printOSD("  - Creating keyboard device\n");
	result = IDirectInput_CreateDevice(lpDI, &GUID_SysKeyboard, &lpDIDKeyboard, NULL);
	if (result != DI_OK) {
		ShowDInputErrorBox("Failed creating keyboard device", result);
		UninitDirectInput();
		return TRUE;
	}

	result = IDirectInputDevice_SetDataFormat(lpDIDKeyboard, &c_dfDIKeyboard);
	if (result != DI_OK) {
		ShowDInputErrorBox("Failed setting keyboard data format", result);
		UninitDirectInput();
		return TRUE;
	}

#ifdef USE_BUFFERED_KEYBOARD
	// set up keyboard for buffered IO
	dipdw.diph.dwSize = sizeof(DIPROPDWORD);
	dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
	dipdw.diph.dwObj = 0;
	dipdw.diph.dwHow = DIPH_DEVICE;
	dipdw.dwData = INPUT_BUFFER_SIZE;

	result = IDirectInputDevice_SetProperty(lpDIDKeyboard, DIPROP_BUFFERSIZE, &dipdw.diph);
	if (result != DI_OK) {
		ShowDInputErrorBox("Failed setting keyboard buffering", result);
		UninitDirectInput();
		return TRUE;
	}
#endif

	// create the mouse device
	printOSD("  - Creating mouse device\n");
	result = IDirectInput_CreateDevice(lpDI, &GUID_SysMouse, &lpDIDMouse, NULL);
	if (result != DI_OK) {
		ShowDInputErrorBox("Failed creating mouse device", result);
		UninitDirectInput();
		return TRUE;
	}

	result = IDirectInputDevice_SetDataFormat(lpDIDMouse, &c_dfDIMouse);
	if (result != DI_OK) {
		ShowDInputErrorBox("Failed setting mouse data format", result);
		UninitDirectInput();
		return TRUE;
	}


	bDInputInited = TRUE;
	return FALSE;
}


//
// UninitDirectInput() -- clean up DirectInput
//
static void UninitDirectInput(void)
{
	if (bDInputInited) printOSD("Uninitializing DirectInput...\n");

	AcquireInputDevices(0);

	if (lpDIDKeyboard) {
		printOSD("  - Releasing keyboard device\n");
		IDirectInputDevice_Release(lpDIDKeyboard);
		lpDIDKeyboard = NULL;
	}

	if (lpDIDMouse) {
		printOSD("  - Releasing mouse device\n");
		IDirectInputDevice_Release(lpDIDMouse);
		lpDIDMouse = NULL;
	}

	if (lpDI) {
		printOSD("  - Releasing DirectInput object\n");
		IDirectInput_Release(lpDI);
		lpDI = NULL;
	}

	if (hDInputDLL) {
		printOSD("  - Unloading DINPUT.DLL\n");
		FreeLibrary(hDInputDLL);
		hDInputDLL = NULL;
	}

	bDInputInited = FALSE;
}


//
// AcquireInputDevices() -- (un)acquires the input devices
//
static void AcquireInputDevices(char acquire)
{
	DWORD flags;

	if (!bDInputInited) return;

	if (acquire) {
		if (lpDIDKeyboard) {
			IDirectInputDevice_SetCooperativeLevel(lpDIDKeyboard, hWindow, DISCL_FOREGROUND|DISCL_NONEXCLUSIVE);
			IDirectInputDevice_Acquire(lpDIDKeyboard);
		}

		if (lpDIDMouse) {
			if (fullscreen || mousegrab) flags = DISCL_FOREGROUND|DISCL_EXCLUSIVE;
			else flags = DISCL_FOREGROUND|DISCL_NONEXCLUSIVE;
			IDirectInputDevice_SetCooperativeLevel(lpDIDMouse, hWindow, flags);
			IDirectInputDevice_Acquire(lpDIDMouse);
		}
	} else {
		if (lpDIDKeyboard) {
			IDirectInputDevice_SetCooperativeLevel(lpDIDKeyboard, hWindow, DISCL_FOREGROUND|DISCL_NONEXCLUSIVE);
			IDirectInputDevice_Unacquire(lpDIDKeyboard);
		}

		if (lpDIDMouse) {
			IDirectInputDevice_SetCooperativeLevel(lpDIDMouse, hWindow, DISCL_FOREGROUND|DISCL_NONEXCLUSIVE);
			IDirectInputDevice_Unacquire(lpDIDMouse);
		}
	}

	// reset mouse and keypresses
	mousex = 0;
	mousey = 0;
	mouseb = 0;

	memset(keystatus, 0, sizeof(keystatus));
	memset(keyfifo, 0, sizeof(keyfifo));
	keyfifoplc = 0;
	keyfifoend = 0;
	lastKeyDown = 0;
}


//
// ProcessInputDevices() -- processes the input devices
//
static void ProcessInputDevices(void)
{
	DIMOUSESTATE dims;
	DWORD i;
	HRESULT result;
#ifdef USE_BUFFERED_KEYBOARD
	DIDEVICEOBJECTDATA didod[INPUT_BUFFER_SIZE];
	DWORD dwElements;
#else
	BYTE *ks, *oks, a,b;
#endif

	if (lpDIDKeyboard) {
#ifdef USE_BUFFERED_KEYBOARD
		// buffered
		// 
		dwElements = INPUT_BUFFER_SIZE;
		result = IDirectInputDevice_GetDeviceData(lpDIDKeyboard, sizeof(DIDEVICEOBJECTDATA), (LPDIDEVICEOBJECTDATA)&didod, &dwElements, 0);
		if (result == DIERR_INPUTLOST) {
			IDirectInputDevice_Acquire(lpDIDKeyboard);
			result = IDirectInputDevice_GetDeviceData(lpDIDKeyboard, sizeof(DIDEVICEOBJECTDATA), (LPDIDEVICEOBJECTDATA)&didod, &dwElements, 0);
		}

		if (result == DI_OK) {
			// process the key events
			for (i=0; i<dwElements; i++) {
				SetKey(didod[i].dwOfs, (didod[i].dwData & 0x80) == 0x80);

				if ((lastKeyDown == didod[i].dwOfs) && !(didod[i].dwData & 0x80))
					lastKeyDown = 0;
				else if (didod[i].dwData & 0x80) {
					lastKeyDown = didod[i].dwOfs;
					lastKeyTime = timeGetTime() + 250;
				}
			}

			if (lastKeyDown > 0) {
				if (timeGetTime() >= lastKeyTime) {
					SetKey(lastKeyDown, 1);
					lastKeyTime = timeGetTime() + 30;
				}
			}
//		} else {
//			printOSD("Failed querying keyboard device data: %s\n", GetDInputError(result));
		}
#else
		// unbuffered
		//
		ks = oks = keys;
		if (!keysofs) ks+=256;
		else oks+=256;
		keysofs = !keysofs;
	
		result = IDirectInputDevice_GetDeviceState(lpDIDKeyboard, sizeof(BYTE)*256, ks);
		if (result == DIERR_INPUTLOST) {
			IDirectInputDevice_Acquire(lpDIDKeyboard);
			result = IDirectInputDevice_GetDeviceState(lpDIDKeyboard, sizeof(BYTE)*256, ks);
		}
		
		if (result == DI_OK) {
			for (i=0; i<256; i++) {
				a = ks[i]&0x80;
				b = oks[i]&0x80;
				if (a == b) continue;	// no change in state
				
				// if the last key pressed was this current one,
				// and the key has just been released, stop repeating
				if ((i == lastKeyDown) && b && !a)
					lastKeyDown = 0;

				SetKey(i, a == 0x80);
				if (lastKeyDown) continue;
				if (a) {
					lastKeyDown = i;
					lastKeyTime = timeGetTime() + 250;
				}
			}

			if (lastKeyDown > 0) {
				if (timeGetTime() >= lastKeyTime) {
					SetKey(lastKeyDown, 1);
					lastKeyTime = timeGetTime() + 30;
				}
			}
		}
#endif
	}

	if (lpDIDMouse && moustat) {
		result = IDirectInputDevice_GetDeviceState(lpDIDMouse, sizeof(dims), &dims);
		if (result == DIERR_INPUTLOST) {
			IDirectInputDevice_Acquire(lpDIDMouse);
			result = IDirectInputDevice_GetDeviceState(lpDIDMouse, sizeof(dims), &dims);
		}

		if (result == DI_OK) {
			// update the mouse statuses
			//   5,4,3,2,1,0 = wheelup,wheeldown,thumb,middle,right,left
			mousex = (short)dims.lX;
			mousey = (short)dims.lY;
			mouseb = 0;
			if (dims.lZ < 0) mouseb |= 16;
			else if (dims.lZ > 0) mouseb |= 32;
			if (dims.rgbButtons[0] & 0x80) mouseb |= 1;
			if (dims.rgbButtons[1] & 0x80) mouseb |= 2;
			if (dims.rgbButtons[2] & 0x80) mouseb |= 4;
			if (dims.rgbButtons[3] & 0x80) mouseb |= 8;
//		} else {
//			printOSD("Failed querying mouse device data: %s\n", GetDInputError(result));
		}
	}
}


//
// UninitDIB() -- clean up the DIB renderer
//
static void UninitDIB(void)
{
	AcquireInputDevices(0);

	if (desktoppal)
		RestoreSystemColours();

	if (hPalette) {
		DeleteObject(hPalette);
		hPalette = NULL;
	}

	if (hDCSection) {
		DeleteDC(hDCSection);
		hDCSection = NULL;
	}

	if (hDIBSection) {
		DeleteObject(hDIBSection);
		hDIBSection = NULL;
	}

	if (hDC) {
		ReleaseDC(hWindow, hDC);
		hDC = NULL;
	}
}


//
// CreateAppWindow() -- create the application window
//
static BOOL CreateAppWindow(int width, int height, int fs)
{
	TCHAR msg[1024];
	RECT rect;
	int w, h, x, y, i;
	struct binfo {
		BITMAPINFOHEADER header;
		RGBQUAD colours[256];
	} dibsect;
	VOID *pixels;

	HRESULT result;
	DDSURFACEDESC ddsd;


	if (!hWindow) {
		hWindow = CreateWindowEx(
			0,
			"buildapp",
			apptitle,
			WINDOW_STYLE,
			CW_USEDEFAULT,
			CW_USEDEFAULT,
			320,
			200,
			NULL,
			NULL,
			hInstance,
			0);
		if (!hWindow) {
			ShowErrorBox("Unable to create window");
			return TRUE;
		}
	}

	// if we're coming out of fullscreen mode, reset the current display mode
	// so the window resizing doesn't get screwed up (???)
//	if (fullscreen && !fs) {
		ReleaseSurfaces();

		if (fullscreen) {
			// restore previous display mode and set to normal cooperative level
			result = IDirectDraw_RestoreDisplayMode(lpDD);
			if (result != DD_OK) {
				ShowDDrawErrorBox("Error restoring display mode", result);
				UninitDirectDraw();
				return TRUE;
			}

			result = IDirectDraw_SetCooperativeLevel(lpDD, hWindow, DDSCL_NORMAL);
			if (result != DD_OK) {
				ShowDDrawErrorBox("Error setting cooperative level", result);
				UninitDirectDraw();
				return TRUE;
			}
		}
//	}

	// resize the window
	if (!fs) {
		rect.left = 0;
		rect.top = 0;
		rect.right = width-1;
		rect.bottom = height-1;
		AdjustWindowRect(&rect, WINDOW_STYLE, FALSE);

		w = (rect.right - rect.left);
		h = (rect.bottom - rect.top);
		x = (desktopxdim - w) / 2;
		y = (desktopydim - h) / 2;

		SetWindowPos(hWindow, HWND_TOP, x, y, w, h, 0);
	} else {
		SetWindowPos(hWindow, HWND_TOP, 0, 0, width, height, 0);
	}

	SetFocus(hWindow);
	ShowWindow(hWindow, SW_SHOWNORMAL);
	UpdateWindow(hWindow);
	SetForegroundWindow(hWindow);

	// fullscreen?
	if (!fs) {
		// no, use DIB section

		if (!hDC) {
			hDC = GetDC(hWindow);
			if (!hDC) {
				ShowErrorBox("Error getting device context");
				return TRUE;
			}
		}

		if (hDCSection) {
			DeleteDC(hDCSection);
			hDCSection = NULL;
		}

		// destroy the previous DIB section if it existed
		if (hDIBSection) {
			DeleteObject(hDIBSection);
			hDIBSection = NULL;
		}

		// create the new DIB section
		memset(&dibsect, 0, sizeof(dibsect));
		dibsect.header.biSize = sizeof(dibsect.header);
		dibsect.header.biWidth = width;
		dibsect.header.biHeight = -height;
		dibsect.header.biPlanes = 1;
		dibsect.header.biBitCount = 8;
		dibsect.header.biCompression = BI_RGB;
		dibsect.header.biClrUsed = 256;
		dibsect.header.biClrImportant = 256;
		for (i=0; i<256; i++) {
			dibsect.colours[i].rgbBlue = curPalette[i].peBlue;
			dibsect.colours[i].rgbGreen = curPalette[i].peGreen;
			dibsect.colours[i].rgbRed = curPalette[i].peRed;
		}

		hDIBSection = CreateDIBSection(hDC, (BITMAPINFO *)&dibsect, DIB_RGB_COLORS, &pixels, NULL, 0);
		if (!hDIBSection) {
			ReleaseDC(hWindow, hDC);
			hDC = NULL;

			ShowErrorBox("Error creating DIB section");
			return TRUE;
		}

		memset(pixels, 0, width*height);

		// create a compatible memory DC
		hDCSection = CreateCompatibleDC(hDC);
		if (!hDCSection) {
			ReleaseDC(hWindow, hDC);
			hDC = NULL;

			ShowErrorBox("Error creating compatible DC");
			return TRUE;
		}

		// select the DIB section into the memory DC
		if (!SelectObject(hDCSection, hDIBSection)) {
			ReleaseDC(hWindow, hDC);
			hDC = NULL;
			DeleteDC(hDCSection);
			hDCSection = NULL;

			ShowErrorBox("Error creating compatible DC");
			return TRUE;
		}

		xres = width;
		yres = height;

		bytesperline = width;
		frameplace = (long)pixels;

		imageSize = bytesperline*yres;

		fullscreen = fs;
	} else {
		// yes, set up DirectDraw

		// clean up after the DIB renderer if it was being used
		UninitDIB();

		if (!bDDrawInited) {
			DestroyWindow(hWindow);
			hWindow = NULL;
			return TRUE;
		}

		ReleaseSurfaces();	// discard previous mode's surfaces

		// set exclusive cooperative level
		result = IDirectDraw_SetCooperativeLevel(lpDD, hWindow, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Error setting cooperative level", result);
			UninitDirectDraw();
			return TRUE;
		}

		result = IDirectDraw_RestoreDisplayMode(lpDD);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Error restoring display mode", result);
			UninitDirectDraw();
			return TRUE;
		}

		result = IDirectDraw_SetDisplayMode(lpDD, width, height, 8);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Error setting display mode", result);
			UninitDirectDraw();
			return TRUE;
		}

		// now create the DirectDraw surfaces
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount = 1;

		printOSD("  - Creating primary surface\n");
		result = IDirectDraw_CreateSurface(lpDD, &ddsd, &lpDDSPrimary, NULL);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Failure creating primary surface", result);
			UninitDirectDraw();
			return TRUE;
		}

		ZeroMemory(&ddsd.ddsCaps, sizeof(ddsd.ddsCaps));
		ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;

		printOSD("  - Getting back buffer\n");
		result = IDirectDrawSurface_GetAttachedSurface(lpDDSPrimary, &ddsd.ddsCaps, &lpDDSBack);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Failure fetching back-buffer surface", result);
			UninitDirectDraw();
			return TRUE;
		}

		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth = width;
		ddsd.dwHeight = height;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

		printOSD("  - Creating offscreen surface\n");
		result = IDirectDraw_CreateSurface(lpDD, &ddsd, &lpDDSOffscreen, NULL);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Failure creating offscreen surface", result);
			UninitDirectDraw();
			return TRUE;
		}

		// attach a palette to the primary surface
		printOSD("  - Creating palette\n");
		result = IDirectDraw_CreatePalette(lpDD, DDPCAPS_8BIT | DDPCAPS_ALLOW256, curPalette, &lpDDPalette, NULL);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Failure creating palette", result);
			UninitDirectDraw();
			return TRUE;
		}

		result = IDirectDrawSurface_SetPalette(lpDDSPrimary, lpDDPalette);
		if (result != DD_OK) {
			ShowDDrawErrorBox("Failure setting palette", result);
			UninitDirectDraw();
			return TRUE;
		}

		memset(&ddsd, 0, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);

		result = IDirectDrawSurface_Lock(lpDDSOffscreen, NULL, &ddsd, DDLOCK_NOSYSLOCK | DDLOCK_WAIT, NULL);
		if (result != DD_OK) {
			printOSD("Failed locking offscreen surface: %s\n", GetDDrawError(result));
			return -1;
		}

		xres = width;
		yres = height;

		bytesperline = ddsd.lPitch;
		frameplace = (long)ddsd.lpSurface;

		imageSize = bytesperline*yres;

		fullscreen = fs;
	}

	return FALSE;
}


//
// DestroyAppWindow() -- destroys the application window
//
static void DestroyAppWindow(void)
{
	UninitDirectDraw();
	UninitDIB();

	if (hWindow) {
		DestroyWindow(hWindow);
		hWindow = NULL;
	}
}


//
// GetDDrawError() -- stinking huge list of error messages since MS didn't want to include
//   them in the DLL
//
static const char * GetDDrawError(HRESULT code)
{
	switch (code) {
		case DD_OK:
			return "The request completed successfully.";
		case DDERR_ALREADYINITIALIZED:
			return "The object has already been initialized.";
		case DDERR_BLTFASTCANTCLIP:
			return "A DirectDrawClipper object is attached to a source surface that has passed into "
				"a call to the IDirectDrawSurface4::BltFast method.";
		case DDERR_CANNOTATTACHSURFACE:
			return "A surface cannot be attached to another requested surface.";
		case DDERR_CANNOTDETACHSURFACE:
			return "A surface cannot be detached from another requested surface.";
		case DDERR_CANTCREATEDC:
			return "Windows can not create any more device contexts (DCs), or a DC was requested for "
				"a palette-indexed surface when the surface had no palette and the display mode "
				"was not palette-indexed (in this case DirectDraw cannot select a proper palette "
				"into the DC).";
		case DDERR_CANTDUPLICATE:
			return "Primary and 3-D surfaces, or surfaces that are implicitly created, cannot be "
				"duplicated.";
		case DDERR_CANTLOCKSURFACE:
			return "Access to this surface is refused because an attempt was made to lock the primary "
				"surface without DCI support.";
		case DDERR_CANTPAGELOCK:
			return "An attempt to page lock a surface failed. Page lock will not work on a "
				"display-memory surface or an emulated primary surface.";
		case DDERR_CANTPAGEUNLOCK:
			return "An attempt to page unlock a surface failed. Page unlock will not work on a "
				"display-memory surface or an emulated primary surface.";
		case DDERR_CLIPPERISUSINGHWND:
			return "An attempt was made to set a clip list for a DirectDrawClipper object that is "
				"already monitoring a window handle.";
		case DDERR_COLORKEYNOTSET:
			return "No source color key is specified for this operation.";
		case DDERR_CURRENTLYNOTAVAIL:
			return "No support is currently available.";
		case DDERR_DCALREADYCREATED:
			return "A device context (DC) has already been returned for this surface. Only one DC "
				"can be retrieved for each surface.";
		case DDERR_DEVICEDOESNTOWNSURFACE:
			return "Surfaces created by one DirectDraw device cannot be used directly by another "
				"DirectDraw device.";
		case DDERR_DIRECTDRAWALREADYCREATED:
			return "A DirectDraw object representing this driver has already been created for this "
				"process.";
		case DDERR_EXCEPTION:
			return "An exception was encountered while performing the requested operation.";
		case DDERR_EXCLUSIVEMODEALREADYSET:
			return "An attempt was made to set the cooperative level when it was already set to "
				"exclusive.";
		case DDERR_EXPIRED:
			return "The data has expired and is therefore no longer valid.";
		case DDERR_GENERIC:
			return "There is an undefined error condition.";
		case DDERR_HEIGHTALIGN:
			return "The height of the provided rectangle is not a multiple of the required "
				"alignment.";
		case DDERR_HWNDALREADYSET:
			return "The DirectDraw cooperative level window handle has already been set. It cannot be "
				"reset while the process has surfaces or palettes created.";
		case DDERR_HWNDSUBCLASSED:
			return "DirectDraw is prevented from restoring state because the DirectDraw cooperative "
				"level window handle has been subclassed.";
		case DDERR_IMPLICITLYCREATED:
			return "The surface cannot be restored because it is an implicitly created surface.";
		case DDERR_INCOMPATIBLEPRIMARY:
			return "The primary surface creation request does not match with the existing primary "
				"surface.";
		case DDERR_INVALIDCAPS:
			return "One or more of the capability bits passed to the callback function are "
				"incorrect.";
		case DDERR_INVALIDCLIPLIST:
			return "DirectDraw does not support the provided clip list.";
		case DDERR_INVALIDDIRECTDRAWGUID:
			return "The globally unique identifier (GUID) passed to the DirectDrawCreate function is "
				"not a valid DirectDraw driver identifier.";
		case DDERR_INVALIDMODE:
			return "DirectDraw does not support the requested mode.";
		case DDERR_INVALIDOBJECT:
			return "DirectDraw received a pointer that was an invalid DirectDraw object.";
		case DDERR_INVALIDPARAMS:
			return "One or more of the parameters passed to the method are incorrect.";
		case DDERR_INVALIDPIXELFORMAT:
			return "The pixel format was invalid as specified.";
		case DDERR_INVALIDPOSITION:
			return "The position of the overlay on the destination is no longer legal.";
		case DDERR_INVALIDRECT:
			return "The provided rectangle was invalid.";
		case DDERR_INVALIDSTREAM:
			return "The specified stream contains invalid data.";
		case DDERR_INVALIDSURFACETYPE:
			return "The requested operation could not be performed because the surface was of the "
				"wrong type.";
		case DDERR_LOCKEDSURFACES:
			return "One or more surfaces are locked, causing the failure of the requested operation.";
		case DDERR_MOREDATA:
			return "There is more data available than the specified buffer size can hold.";
		case DDERR_NO3D:
			return "No 3-D hardware or emulation is present.";
		case DDERR_NOALPHAHW:
			return "No alpha acceleration hardware is present or available, causing the failure of "
				"the requested operation.";
		case DDERR_NOBLTHW:
			return "No blitter hardware is present.";
		case DDERR_NOCLIPLIST:
			return "No clip list is available.";
		case DDERR_NOCLIPPERATTACHED:
			return "No DirectDrawClipper object is attached to the surface object.";
		case DDERR_NOCOLORCONVHW:
			return "The operation cannot be carried out because no color-conversion hardware is "
				"present or available.";
		case DDERR_NOCOLORKEY:
			return "The surface does not currently have a color key.";
		case DDERR_NOCOLORKEYHW:
			return "The operation cannot be carried out because there is no hardware support for the "
				"destination color key.";
		case DDERR_NOCOOPERATIVELEVELSET:
			return "A create function is called without the IDirectDraw4::SetCooperativeLevel method "
				"being called.";
		case DDERR_NODC:
			return "No DC has ever been created for this surface.";
		case DDERR_NODDROPSHW:
			return "No DirectDraw raster operation (ROP) hardware is available.";
		case DDERR_NODIRECTDRAWHW:
			return "Hardware-only DirectDraw object creation is not possible; the driver does not "
				"support any hardware.";
		case DDERR_NODIRECTDRAWSUPPORT:
			return "DirectDraw support is not possible with the current display driver.";
		case DDERR_NOEMULATION:
			return "Software emulation is not available.";
		case DDERR_NOEXCLUSIVEMODE:
			return "The operation requires the application to have exclusive mode, but the application "
				"does not have exclusive mode.";
		case DDERR_NOFLIPHW:
			return "Flipping visible surfaces is not supported.";
		case DDERR_NOFOCUSWINDOW:
			return "An attempt was made to create or set a device window without first setting the "
				"focus window.";
		case DDERR_NOGDI:
			return "No GDI is present.";
		case DDERR_NOHWND:
			return "Clipper notification requires a window handle, or no window handle has been "
				"previously set as the cooperative level window handle.";
		case DDERR_NOMIPMAPHW:
			return "The operation cannot be carried out because no mipmap capable texture mapping "
				"hardware is present or available.";
		case DDERR_NOMIRRORHW:
			return "The operation cannot be carried out because no mirroring hardware is present or "
				"available.";
		case DDERR_NONONLOCALVIDMEM:
			return "An attempt was made to allocate non-local video memory from a device that does "
				"not support non-local video memory.";
		case DDERR_NOOPTIMIZEHW:
			return "The device does not support optimized surfaces.";
		case DDERR_NOOVERLAYDEST:
			return "The IDirectDrawSurface4::GetOverlayPosition method is called on an overlay that "
				"the IDirectDrawSurface4::UpdateOverlay method has not been called on to establish "
				"a destination.";
		case DDERR_NOOVERLAYHW:
			return "The operation cannot be carried out because no overlay hardware is present or "
				"available.";
		case DDERR_NOPALETTEATTACHED:
			return "No palette object is attached to this surface.";
		case DDERR_NOPALETTEHW:
			return "There is no hardware support for 16- or 256-color palettes.";
		case DDERR_NORASTEROPHW:
			return "The operation cannot be carried out because no appropriate raster operation "
				"hardware is present or available.";
		case DDERR_NOROTATIONHW:
			return "The operation cannot be carried out because no rotation hardware is present or "
				"available.";
		case DDERR_NOSTRETCHHW:
			return "The operation cannot be carried out because there is no hardware support for "
				"stretching.";
		case DDERR_NOT4BITCOLOR:
			return "The DirectDrawSurface object is not using a 4-bit color palette and the requested "
				"operation requires a 4-bit color palette.";
		case DDERR_NOT4BITCOLORINDEX:
			return "The DirectDrawSurface object is not using a 4-bit color index palette and the "
				"requested operation requires a 4-bit color index palette.";
		case DDERR_NOT8BITCOLOR:
			return "The DirectDrawSurface object is not using an 8-bit color palette and the "
				"requested operation requires an 8-bit color palette.";
		case DDERR_NOTAOVERLAYSURFACE:
			return "An overlay component is called for a non-overlay surface.";
		case DDERR_NOTEXTUREHW:
			return "The operation cannot be carried out because no texture-mapping hardware is present "
				"or available.";
		case DDERR_NOTFLIPPABLE:
			return "An attempt has been made to flip a surface that cannot be flipped.";
		case DDERR_NOTFOUND:
			return "The requested item was not found.";
		case DDERR_NOTINITIALIZED:
			return "An attempt was made to call an interface method of a DirectDraw object created by "
				"CoCreateInstance before the object was initialized.";
		case DDERR_NOTLOADED:
			return "The surface is an optimized surface, but it has not yet been allocated any memory.";
		case DDERR_NOTLOCKED:
			return "An attempt is made to unlock a surface that was not locked.";
		case DDERR_NOTPAGELOCKED:
			return "An attempt is made to page unlock a surface with no outstanding page locks.";
		case DDERR_NOTPALETTIZED:
			return "The surface being used is not a palette-based surface.";
		case DDERR_NOVSYNCHW:
			return "The operation cannot be carried out because there is no hardware support for "
				"vertical blank synchronized operations.";
		case DDERR_NOZBUFFERHW:
			return "The operation to create a z-buffer in display memory or to perform a blit using "
				"a z-buffer cannot be carried out because there is no hardware support for z-buffers.";
		case DDERR_NOZOVERLAYHW:
			return "The overlay surfaces cannot be z-layered based on the z-order because the hardware "
				"does not support z-ordering of overlays.";
		case DDERR_OUTOFCAPS:
			return "The hardware needed for the requested operation has already been allocated.";
		case DDERR_OUTOFMEMORY:
			return "DirectDraw does not have enough memory to perform the operation.";
		case DDERR_OUTOFVIDEOMEMORY:
			return "DirectDraw does not have enough display memory to perform the operation.";
		case DDERR_OVERLAPPINGRECTS:
			return "Operation could not be carried out because the source and destination rectangles are "
				"on the same surface and overlap each other.";
		case DDERR_OVERLAYCANTCLIP:
			return "The hardware does not support clipped overlays.";
		case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
			return "An attempt was made to have more than one color key active on an overlay.";
		case DDERR_OVERLAYNOTVISIBLE:
			return "The IDirectDrawSurface4::GetOverlayPosition method is called on a hidden overlay.";
		case DDERR_PALETTEBUSY:
			return "Access to this palette is refused because the palette is locked by another thread.";
		case DDERR_PRIMARYSURFACEALREADYEXISTS:
			return "This process has already created a primary surface.";
		case DDERR_REGIONTOOSMALL:
			return "The region passed to the IDirectDrawClipper::GetClipList method is too small.";
		case DDERR_SURFACEALREADYATTACHED:
			return "An attempt was made to attach a surface to another surface to which it is already "
				"attached.";
		case DDERR_SURFACEALREADYDEPENDENT:
			return "An attempt was made to make a surface a dependency of another surface to which it "
				"is already dependent.";
		case DDERR_SURFACEBUSY:
			return "Access to the surface is refused because the surface is locked by another thread.";
		case DDERR_SURFACEISOBSCURED:
			return "Access to the surface is refused because the surface is obscured.";
		case DDERR_SURFACELOST:
			return "Access to the surface is refused because the surface memory is gone. Call the "
				"IDirectDrawSurface4::Restore method on this surface to restore the memory "
				"associated with it.";
		case DDERR_SURFACENOTATTACHED:
			return "The requested surface is not attached.";
		case DDERR_TOOBIGHEIGHT:
			return "The height requested by DirectDraw is too large.";
		case DDERR_TOOBIGSIZE:
			return "The size requested by DirectDraw is too large. However, the individual height and "
				"width are valid sizes.";
		case DDERR_TOOBIGWIDTH:
			return "The width requested by DirectDraw is too large.";
		case DDERR_UNSUPPORTED:
			return "The operation is not supported.";
		case DDERR_UNSUPPORTEDFORMAT:
			return "The pixel format requested is not supported by DirectDraw.";
		case DDERR_UNSUPPORTEDMASK:
			return "The bitmask in the pixel format requested is not supported by DirectDraw.";
		case DDERR_UNSUPPORTEDMODE:
			return "The display is currently in an unsupported mode.";
		case DDERR_VERTICALBLANKINPROGRESS:
			return "A vertical blank is in progress.";
		case DDERR_VIDEONOTACTIVE:
			return "The video port is not active.";
		case DDERR_WASSTILLDRAWING:
			return "The previous blit operation that is transferring information to or from this "
				"surface is incomplete.";
		case DDERR_WRONGMODE:
			return "This surface cannot be restored because it was created in a different mode.";
		case DDERR_XALIGN:
			return "The provided rectangle was not horizontally aligned on a required boundary.";
		default:
			break;
	}
	return "Unknown error";
}


//
// GetDInputError() -- stinking huge list of error messages since MS didn't want to include
//   them in the DLL
//
static const char * GetDInputError(HRESULT code)
{
	switch (code) {
		case DI_OK:
			return "The operation completed successfully.";
		case DI_BUFFEROVERFLOW:
			return "The device buffer overflowed and some input was lost.";
		case DI_DOWNLOADSKIPPED:
			return "The parameters of the effect were successfully updated, but the effect could not "
				"be downloaded because the associated device was not acquired in exclusive mode.";
		case DI_EFFECTRESTARTED:
			return "The effect was stopped, the parameters were updated, and the effect was restarted.";
		case DI_POLLEDDEVICE:
			return "The device is a polled device. As a result, device buffering will not collect any "
				"data and event notifications will not be signaled until the "
				"IDirectInputDevice2::Poll method is called.";
		case DI_TRUNCATED:
			return "The parameters of the effect were successfully updated, but some of them were "
				"beyond the capabilities of the device and were truncated to the nearest supported "
				"value.";
		case DI_TRUNCATEDANDRESTARTED:
			return "See DI_EFFECTRESTARTED | DI_TRUNCATED.";
		case DIERR_ACQUIRED:
			return "The operation cannot be performed while the device is acquired.";
		case DIERR_ALREADYINITIALIZED:
			return "This object is already initialized.";
		case DIERR_BADDRIVERVER:
			return "The object could not be created due to an incompatible driver version or mismatched "
				"or incomplete driver components.";
		case DIERR_BETADIRECTINPUTVERSION:
			return "The application was written for an unsupported prerelease version of DirectInput.";
		case DIERR_DEVICEFULL:
			return "The device is full.";
		case DIERR_DEVICENOTREG:
			return "The device or device instance is not registered with DirectInput.";
		case DIERR_EFFECTPLAYING:
			return "The parameters were updated in memory but were not downloaded to the device because "
				"the device does not support updating an effect while it is still playing.";
		case DIERR_HASEFFECTS:
			return "The device cannot be reinitialized because there are still effects attached to it.";
		case DIERR_GENERIC:
			return "An undetermined error occurred inside the DirectInput subsystem.";
		case DIERR_HANDLEEXISTS:
			return "The device already has an event notification associated with it.";
		case DIERR_INCOMPLETEEFFECT:
			return "The effect could not be downloaded because essential information is missing. For "
				"example, no axes have been associated with the effect, or no type-specific "
				"information has been supplied.";
		case DIERR_INPUTLOST:
			return "Access to the input device has been lost. It must be reacquired.";
		case DIERR_INVALIDPARAM:
			return "An invalid parameter was passed to the returning function, or the object was not in "
				"a state that permitted the function to be called.";
		case DIERR_MOREDATA:
			return "Not all the requested information fitted into the buffer.";
		case DIERR_NOAGGREGATION:
			return "This object does not support aggregation.";
		case DIERR_NOINTERFACE:
			return "The specified interface is not supported by the object.";
		case DIERR_NOTACQUIRED:
			return "The operation cannot be performed unless the device is acquired.";
		case DIERR_NOTBUFFERED:
			return "The device is not buffered. Set the DIPROP_BUFFERSIZE property to enable buffering.";
		case DIERR_NOTDOWNLOADED:
			return "The effect is not downloaded.";
		case DIERR_NOTEXCLUSIVEACQUIRED:
			return "The operation cannot be performed unless the device is acquired in DISCL_EXCLUSIVE "
				"mode.";
		case DIERR_NOTFOUND:
			return "The requested object does not exist.";
		case DIERR_NOTINITIALIZED:
			return "This object has not been initialized.";
		case DIERR_OLDDIRECTINPUTVERSION:
			return "The application requires a newer version of DirectInput.";
		case DIERR_OUTOFMEMORY:
			return "The DirectInput subsystem couldn't allocate sufficient memory to complete the call.";
		case DIERR_UNSUPPORTED:
			return "The function called is not supported at this time.";
		case E_PENDING:
			return "Data is not yet available.";
		default:
			break;
	}
	return "Unknown error";
}


//
// GetWindowsErrorMsg() -- gives a pointer to a static buffer containing the Windows error message
//
static const LPTSTR GetWindowsErrorMsg(DWORD code)
{
	static TCHAR lpMsgBuf[1024];

	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL, code,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR)lpMsgBuf, 1024, NULL);

	return lpMsgBuf;
}


//
// SaveSystemColours() -- save the Windows-reserved colours
//
static void SaveSystemColours(void)
{
	int i;

	if (system_colours_saved) return;

	for (i=0; i<NUM_SYS_COLOURS; i++)
		syscolours[i] = GetSysColor(syscolouridx[i]);

	system_colours_saved = 1;
}


//
// RestoreSystemColours() -- restore the Windows-reserved colours
//
static void RestoreSystemColours(void)
{
	if (!system_colours_saved) return;

	SetSystemPaletteUse(hDC, SYSPAL_STATIC);
	SetSysColors(NUM_SYS_COLOURS, syscolouridx, syscolours);
}

